/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __PIN_H__
#define __PIN_H__

class pin : public symbol
{
  friend net;

  component *comp;
  real Vin; // For display purposes.
  unsigned int hasparent : 1;
  unsigned int modified : 1;

protected:
  virtual void _SetVin(real Vin) {};

public:
  real highcarelevel;
  real lowcarelevel;

  virtual symbolwidget *getwidget();
  void ClearModified() { modified = 0; }
  int GetModified(); 
  void SetVin(real Vin);
  component *GetComponent() { return comp; }
  real GetVin();
  virtual real GetVout() = 0;
  virtual real GetZ() = 0;
  virtual pin *ToPin() { return this; }

  pin(int name, component *comp, attribute *attribs, symbol *parent);
};

class floatpin : public pin
{
  real *ptr;
public:
  virtual real GetVout() { return *ptr; }
  virtual real GetZ() {return 0; }
  floatpin(int name, component *comp, attribute *attribs, symbol *parent, real *ptr)
    :pin(name, comp, attribs, parent)
    { floatpin::ptr = ptr; }
};

class simplepin : public pin
{
protected:
  real Vout, Z;

public:
  virtual real GetVout() { return Vout; }
  virtual real GetZ() {return Z; }
  virtual void SetVout(real Vout) { simplepin::Vout = Vout; }
  virtual void SetZ(real Z) { simplepin::Z = Z; }

  simplepin(int name, component *comp, attribute *attribs, symbol *parent) : pin(name, comp, attribs, parent) {}
};

class net : public simplepin
{
  pinlist *pins;

public:
  net(int name, attribute *attribs, pinlist *pins, symbol *parent, macrocomponent *comp);

  void PropagateDown(int force);
  void PropagateUp();
};

class netconstructor : public symbolconstructor
{
  symbol *_ccreate(macrocomponent *parent);
  list<symbolname> *pinnames;
public:
  netconstructor(int name, attribute *attribs, list<symbolname> *pinnames);
};

class incomppin : public pin
{
  real limit;
  int State;

  virtual list<symbol> *getapparentsymbols();
  virtual symbol *_getlocalsymbol(symbolname *);
protected:
  virtual void _SetVin(real Vin);

public:
  unsigned int care : 1;
  virtual real GetVout() { return NAN; }
  virtual real GetZ() { return INFINITY; }
  virtual int GetState() { return State; }
  virtual void parseattribs(attribute *attribs);

  incomppin(int name, component *comp, attribute *attribs, symbol *parent);
};

class outpin : public pin
{
  int state;
  real hilvl, lolvl, Vout, Z;
  virtual list<symbol> *getapparentsymbols();
  virtual symbol *_getlocalsymbol(symbolname *);
public:
  virtual real GetVout() { return state ? hilvl : lolvl; }
  virtual real GetZ() { return Z; }
  void SetState(int s) { state = s != 0; }
  int GetState() { return state; }
  virtual void parseattribs(attribute *attribs);

  outpin(int name, component *comp, attribute *attribs, symbol *parent, int initstate = 0);
};

class _nullpin : public pin
{
  friend _nullpin *nullpin(attribute *attribs, component *parent);
  static int nullpincount;
  _nullpin(int name, attribute *attribs, component *parent);
public:
  virtual symbolwidget *getwidget();
  virtual real GetVout() { return NAN; };
  virtual real GetZ() { return INFINITY; };
};

#define  word short int
#define  byte unsigned char

//Constantes lectriques
#define V_OL 0.0   //Tension de sortie niveau bas
#define V_OH 5.0   //Tension de sortie niveau haut
#define V_IL_I    1.0   //Tension de seuil trigger entre niveau bas (broches RTCC/OSC1/MCLR}
#define V_IH_I    4.0   //Tension de seuil trigger entre niveau haut
#define V_IL_CMOS 1.5   //Tension de seuil trigger entre niveau bas
#define V_IH_CMOS 3.5   //Tension de seuil trigger entre niveau haut
#define V_IL_TTL  0.8   //Tension de seuil trigger entre niveau bas
#define V_IH_TTL  2.0   //Tension de seuil trigger entre niveau haut
#define Z_LO 25         //Impdance en mode sortie
#define Z_PullUp 20000  //Rsistance de Pull Up
#define Z_HI INFINITY   //Impdance en mode entre


//Structure stockant l'adresse d'un bit en mmoire.
struct bit {
  char *Address;  //Adresse de l'octet contenant le bit
  char Mask;      //Masque pour accder au bit
};

//Macros de gestion des bits
#define DefineBit(b,Adr,Msk) {(b).Address=Adr;(b).Mask=Msk;}
#define GetBit(b) ((*((b).Address))&(b).Mask)
#define SetBit(b) (*((b).Address))|=(b).Mask
#define ClrBit(b) (*((b).Address))&=(char)(~(b).Mask)

class pin_3state : public pin
{
protected:
   real LowLevel,HighLevel;
public:
   bit Tris,Latch,Pin; //Bits de direction, de sortie, d'entre
   bit PullUp;
   // Mthodes
   void _SetVin(real Vin);
   real GetVout();
   real GetZ();
   pin_3state(int name, component *comp, attribute *attribs, symbol *parent) : pin(name,comp,attribs,parent) {};
   void SetTTL(int TTL);
   virtual list<symbol> *getapparentsymbols();
   virtual symbol *_getlocalsymbol(symbolname *);
};

class pin_Output : public pin
{
public:
   bit Latch;
   // Mthodes
   real GetVout();
   real GetZ();
   pin_Output(int name, component *comp, attribute *attribs, symbol *parent,
     char* Address,char Mask);
   virtual list<symbol> *getapparentsymbols();
   virtual symbol *_getlocalsymbol(symbolname *);
};

class pin_Input : public pin
{
protected:
   real LowLevel,HighLevel;
public:
   bit Pin; //Bits de direction, de sortie, d'entre
   // Mthodes
   void _SetVin(real Vin);
   real GetVout();
   real GetZ();
   pin_Input(int name, component *comp, attribute *attribs, symbol *parent,
     char* Address,char Mask);
   virtual list<symbol> *getapparentsymbols();
   virtual symbol *_getlocalsymbol(symbolname *);
};

class pin_Edge : public pin_Input
{
public:
   char Edge;
   char Detect;
   // Mthodes
   void _SetVin(real Vin);
   pin_Edge(int name, component *comp, attribute *attribs, symbol *parent,
     char* Address,char Mask);
   virtual list<symbol> *getapparentsymbols();
   virtual symbol *_getlocalsymbol(symbolname *);
};


#endif
