/*
  DUST: Simulator for circuits containing a SX microcontroller.
  Copyright (C) 1999 Blaise Gassend <blaise at gassend dot com> 
  and Franois-Marie Lefevere <francois-marie dot lefevere at m4x dot org>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "project.h"

#define MAXSYMBOLNAMELENGTH 200

symboltable &_symbtable()
{
  static symboltable  s;// = new symboltable();
  return s;
}

symboltable::symboltable()
{
  byid = (char **) NULL;
  byname = (char **) NULL;
  id = (int *) NULL;
  size = 0;
  allocated = 0;
  anonymouscount = 0;
}

symboltable::~symboltable()
{
  free(id);
  for (int i = 0; i < size; i++)
    free(byid[i]);
  free(byid);
  free(byname);
}

char *symboltable::tostring(list<int> *n)
{
  static char buff[2][MAXSYMBOLNAMELENGTH];

  buff[1][0] = 0;

  if (n == NULL) return "@__emptyname__";

  goto firstloop;
  for (; n != NULL; n = n->next)
  {
    snprintf(buff[1], MAXSYMBOLNAMELENGTH, "%s/", buff[0]);  
firstloop:
    snprintf(buff[0], MAXSYMBOLNAMELENGTH, "%s%s", buff[1], tostring(*n->elem));
  }

  return buff[0];
}

char *symboltable::tostring(int sid)
{
  static char buff[] = "@__anonymous__xxxxxxxxxx";
  
  if (sid >= 0)
  {
    sprintf(buff, "%i", sid);
    return buff;
  }
  if (sid & 1) // Is an anonymous
  {
    if (sid < -anonymouscount * 2 + 1) return (char *) NULL;
    sprintf(buff, "@__anonymous%i__", -(sid - 1) / 2);
    return buff;
  }
  else if (sid < -size * 2) return (char *) NULL;
  else return byid[-sid / 2 - 1];
}

int symboltable::toid(char *s)
{
  if (*s == 0)
  {
    anonymouscount++;
    DEBUG(cerr << "Added " << tostring(-anonymouscount * 2 + 1) << " as ID " << (-anonymouscount * 2 + 1) << "\n";)
    return (-anonymouscount * 2 + 1);
  }

  int first = 0, last = size - 1;
  
  while (first <= last)
  {
    int cur = (first + last) / 2;
    int c = strcmp(s, byname[cur]);
    if (c == 0) return -id[cur] * 2 - 2;
    if (c < 0) last = cur - 1;
    else first = cur + 1;
  }

  if (size >= allocated)
  {
    allocated += ALLOCSTEP;
    byname = (char **) realloc(byname, allocated * sizeof(*byname));
    byid = (char **) realloc(byid, allocated * sizeof(*byid));
    id = (int *) realloc(id, allocated * sizeof(*id));
    if (byname == NULL || byid == NULL || id == NULL)
      error(TYPE_FATALERROR, GetParseLineNum(), GetParseFileName(), "Error allocating memory for symbol table.");
  }

  for (int i = size; i > first; i--)
  {
    id[i] = id[i - 1];
    byname[i] = byname[i - 1];
  }

  byname[first] = strdup(s);
  id[first] = size;
  byid[size] = byname[first];
  size++;
  
  if (byname[first] == NULL) 
    error(TYPE_FATALERROR, GetParseLineNum(), GetParseFileName(), "Error allocating memory for symbol table.");

    DEBUG(cerr << "Added " << tostring(-id[first] * 2 - 2) << " as ID " << (-id[first] * 2 - 2) << "\n";)
  return -id[first] * 2 - 2;
}

list<int> *symboltable::toname(char *s)
{
  int namestart;
  for (namestart = 0; s[namestart] == ' ' || s[namestart] == '\t'; namestart++);
  if (s[namestart] == 0) return (list<int> *) NULL;

  int slashpos;
  for (slashpos = namestart + 1; s[slashpos] != '/' && s[slashpos] != 0; slashpos++);

  int nameend;
  for (nameend = slashpos - 1; s[nameend] == ' ' || s[nameend] == '\t'; nameend--);
  
  char tmp = s[nameend + 1];
  s[nameend + 1] = 0;
  int id = toid(s+namestart);
  s[nameend + 1] = tmp;

  if (s[slashpos] == 0) return new list<int>(new int(id));
  else 
  {
    list<int> *tail = toname(s+slashpos+1);
    if (tail == NULL) return (list<int> *) NULL;
    return new list<int>(new int(id), tail);
  }
}

void symboltable::deletename(list<int> *name)
{
  for (list<int> *l = name; l != NULL; l = l->next)
    delete l->elem;
  delete name;
}
